import binascii
import functools
import logging
import socket
from time import sleep

import getmac
import wakeonlan

HEADER = b'\xA6'
CATEGORY = b'\x00'
CODE0 = b'\x00'
CODE1 = b'\x00'
DATA_CONTROL = b'\x01'


def calculate_checksum(message):
    return bytes([functools.reduce(lambda a, b: a ^ b, list(message))])


def _prepare_message(id, command, data):
    data = command + data
    length = (len(data) + 2).to_bytes(1, byteorder='big')
    message = b''.join([HEADER, id, CATEGORY, CODE0, CODE1, length, DATA_CONTROL, data])
    checksum = calculate_checksum(message)
    return message + checksum


class Client:

    def __init__(self, host, port=5000, timeout=2, wake_delay=7, retries=20, buffer_size=1024, mac=None,
                 wol_target=None):
        self._wake_delay = wake_delay
        self._retries = retries
        self._timeout = timeout
        self._wol_target = wol_target
        self._host = host
        self._port = port
        self._mac = mac if mac else getmac.get_mac_address(ip=self._host, hostname=self._host)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._buffer_size = buffer_size  # Timeout after 5 seconds

    def wake_on_lan(self):
        target_ip = wakeonlan.BROADCAST_IP if self._wol_target is None else self._wol_target
        self._logger.debug('Waking on LAN %s using %s' % (self._mac, target_ip))
        wakeonlan.send_magic_packet(self._mac, ip_address=target_ip)

    def send(self, id: bytes, command: bytes, data: bytes = b''):
        _socket = self._create_and_connect_socket()
        if _socket:
            try:
                message = _prepare_message(id, command, data)
                self._log_debug_request(id, message, data)
                recv = self._call_remote(_socket, message)
                return self._process_response(id, command, recv)
            except socket.timeout:
                self._logger.error("Socket timeout, no response received from the server.")
            except socket.error as e:
                self._logger.error(f"Socket error: {e}")
            finally:
                _socket.close()

    def _call_remote(self, _socket, message):
        _socket.sendall(message)
        recv = _socket.recv(self._buffer_size)
        return recv

    def _create_and_connect_socket(self):
        _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        _socket.settimeout(self._timeout)
        for _ in range(self._retries):
            try:
                _socket.connect((self._host, self._port))
                return _socket
            except socket.error:
                self.wake_on_lan()
                sleep(self._wake_delay)
        self._logger.error("Connection error")
        _socket.close()
        return None

    def _process_response(self, id, command, response_data):

        length = len(response_data)
        checksum_data = response_data[:-1]
        checksum = calculate_checksum(checksum_data)
        header = response_data[0]
        response_id = response_data[1]
        category = response_data[2]
        page = response_data[3]
        response_length = response_data[4] - 2
        control = response_data[5]
        response_command = response_data[6]
        result = response_data[7:7 + response_length - 1]
        response_checksum = response_data[7 + response_length - 1]
        self._log_debug_response(category, checksum, control, header, length, page, response_checksum, response_command,
                                 response_id, response_length, response_data)

        assert header == 0x21
        assert id[0] == response_id
        assert category == 0x00
        assert page == 0x00
        assert control == 0x01
        assert checksum[0] == response_checksum

        if response_command == 0x00:
            response_map = {
                b'\x00': None,
                b'\x01': "Limit Over; The packet was received normally, but the data value was over the upper limit.",
                b'\x02': "Limit Over; The packet was received normally, but the data value was over the lower limit.",
                b'\x03': "Command canceled; The packet was received normally but either the value of data is incorrect or request is not permitted for the current host value.",
                b'\x04': "Parse Error; Received not defined format data or checksum Error."
            }
            error_message = response_map.get(result,
                                             "Unexpected Error; Received unexpected error %s." % binascii.hexlify(
                                                 result))
            if error_message:
                self._logger.error(error_message)
            return result if result == b'\x00' else None
        else:
            assert response_command == command[0], "Command doesn't match. Expected 0x%02x, got 0x%02x" % (
                command[0], response_command)
            self._logger.info("data: %s -> %s" % (binascii.hexlify(result), result.decode('utf-8') if result else ""))
            return result

    def _log_debug_response(self, category, checksum, control, header, length, page, response_checksum,
                            response_command, response_id, response_length, response_data):
        self._logger.debug("     response: %s" % binascii.hexlify(response_data))
        self._logger.debug("header: 0x%02x" % header)
        self._logger.debug("id: 0x%02x" % response_id)
        self._logger.debug("category: 0x%02x" % category)
        self._logger.debug("code0: 0x%02x" % page)
        self._logger.debug("length: %d / %d" % (response_length, length))
        self._logger.debug("control: 0x%02x" % control)
        self._logger.debug("command: 0x%02x" % response_command)
        self._logger.debug("checksum: 0x%02x / %s" % (response_checksum, binascii.hexlify(checksum)))

    def _log_debug_request(self, id, message, data):
        checksum = message[-1]
        self._logger.debug("id: %s" % binascii.hexlify(id))
        self._logger.debug("function: %s" % binascii.hexlify(CODE1))
        self._logger.debug("length: %s" % binascii.hexlify(data[:1]))
        self._logger.debug("data: %s" % binascii.hexlify(data))
        self._logger.debug("checksum: %s" % binascii.hexlify(bytes([checksum])))
        self._logger.debug(" request: %s" % binascii.hexlify(message))



import binascii

from pyamasicp.client import Client

CMD_SET_POWER_STATE = b'\x18'
CMD_GET_POWER_STATE = b'\x19'
CMD_SET_VOLUME = b'\x44'
CMD_GET_VOLUME = b'\x45'
CMD_SET_INPUT_SOURCE = b'\xAC'
CMD_GET_INPUT_SOURCE = b'\xAD'
CMD_IR = b'\xDB'

VAL_POWER_OFF = b'\x01'
VAL_POWER_ON = b'\x02'

IR_POWER = b'\xA0'
IR_MENU = b'\xA1'
IR_INPUT = b'\xA2'
IR_VOL_UP = b'\xA3'
IR_VOL_DOWN = b'\xA4'
IR_MUTE = b'\xA5'
IR_CURSOR_UP = b'\xA6'
IR_CURSOR_DOWN = b'\xA7'
IR_CURSOR_LEFT = b'\xA8'
IR_CURSOR_RIGHT = b'\xA9'
IR_OK = b'\xB1'
IR_RETURN = b'\xB2'
IR_RED = b'\xC1'
IR_GREEN = b'\xC2'
IR_YELLOW = b'\xC3'
IR_BLUE = b'\xC4'
IR_FORMAT = b'\xD1'
IR_INFO = b'\xD2'
IR_BTN_0 = b'\x00'
IR_BTN_1 = b'\x01'
IR_BTN_2 = b'\x02'
IR_BTN_3 = b'\x03'
IR_BTN_4 = b'\x04'
IR_BTN_5 = b'\x05'
IR_BTN_6 = b'\x06'
IR_BTN_7 = b'\x07'
IR_BTN_8 = b'\x08'
IR_BTN_9 = b'\x09'

INPUT_SOURCES = {
    "VIDEO": b'\x00',
    "Display Port": b'\x01',
    "S-VIDEO": b'\x02',
    "COMPONENT": b'\x03',
    "VGA": b'\x05',
    "HDMI 2": b'\x06',
    "Display Port 2": b'\x07',
    "USB 2": b'\x08',
    "Card DVI-D": b'\x09',
    "Display Port 1": b'\x0A',
    "Card OPS": b'\x0B',
    "USB 1": b'\x0C',
    "HDMI 1": b'\x0D',
    "DVI-D": b'\x0E',
    "HDMI 3": b'\x0F',
    "BROWSER": b'\x10',
    "SMARTCMS": b'\x11',
    "DMS (Digital Media Server)": b'\x12',
    "INTERNAL STORAGE": b'\x13',
    "Reserved": b'\x14',
    "Media Player": b'\x16',
    "PDF Player": b'\x17',
    "Custom": b'\x18',
    "HDMI 4": b'\x19',
}


class Commands:

    def __init__(self, client: Client, id=b'\x01'):
        self._client = client
        self._id = id

    def get_power_state(self):
        result = self._client.send(self._id, CMD_GET_POWER_STATE)
        match result:
            case b'\x01':
                return False
            case b'\x02':
                return True
            case _:
                raise CommandException("Unknown power state: %s" % binascii.hexlify(result))

    def set_power_state(self, state: bool):
        self._client.send(self._id, CMD_SET_POWER_STATE, VAL_POWER_ON if state else VAL_POWER_OFF)

    def get_volume(self):
        result = [b for b in self._client.send(self._id, CMD_GET_VOLUME)]
        result.reverse()
        return result

    def set_volume(self, volume=0, output_volume=0):
        self._client.send(self._id, CMD_SET_VOLUME, bytearray([volume, output_volume, 0, 0]))

    def get_input_source(self):
        return [b for b in self._client.send(self._id, CMD_GET_INPUT_SOURCE)]

    def set_input_source(self, input_type=0, input_number=0, osd_style=0, reserved=0):
        self._client.send(self._id, CMD_SET_INPUT_SOURCE, bytearray([input_type, input_number, osd_style, reserved]))

    def ir_command(self, code):
        self._client.send(self._id, CMD_IR, code)


class CommandException(Exception):
    pass



import logging

from pyamasicp import client
from pyamasicp.commands import Commands, IR_VOL_UP, IR_VOL_DOWN, IR_OK
import getmac

logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

HOST = 'iiyama.home'

logger.debug("MAC: %a" % getmac.get_mac_address(ip=HOST, hostname=HOST))

c = Commands(client.Client(HOST, mac="DC:62:94:25:02:B3"))
# c.set_power_state(True)

# c.set_volume(22)
# c.ir_command(IR_OK)
logger.info("Current volume level: %s" % (c.get_volume()))
input_source = c.get_input_source()
logger.info("Current input source: 0x%02X[%d]" % (input_source[0], input_source[1]))

c.set_input_source(0x18)
# c.get_power_state()


# c.send(b'\x01', b'\x18', b'\x01')

# send(b'\x01', b'\xA2', b'\x01')
# send(b'\x01', b'\xAC', b'\x00\x18\x01\x00')

# send(b'\x01', b'\x44', b'\100\20')


# HDMI1:    0x0d[0]
# HDMI2:    0x06[0]
# HDMI3:    0x0F[0]
# Browser:  0x10[0]
# CMS:      0x11[0]
# Files:    0x13[0]
# Media:    0x16[0]
# PDF:      0x17[0]
# Custom:   0x18[0]
